(* D'après le code de Xavier Leroy et Didier Remy :
     http://gallium.inria.fr/~remy/poly/system/camlunix
*)

open Unix

exception Error of string

let error err mes = raise (Error (err ^ ": " ^ mes))

let handle_error f x =
  try f x
  with Error err -> prerr_endline err; exit 2

let retransmit fdin fdout =
  let buffer_size = 4096 in
  let buffer = String.create buffer_size in
  let rec copy() =
    match read fdin buffer 0 buffer_size with
    | 0 -> ()
    | n ->
	ignore (write fdout buffer 0 n);
	copy ()
  in
  copy ()

let default_port = "80"

type regexp =
    { regexp : Str.regexp;
      fields : (int * string option) list; }

let regexp_match r string =
  let get (pos, default) =
    try Str.matched_group pos string
    with Not_found ->
      begin match default with
      | Some s -> s
      | _ -> raise Not_found
      end
  in
  try
    if Str.string_match r.regexp string 0 then
      Some (List.map get r.fields)
    else None
  with Not_found -> None

let host_regexp =
  { regexp = Str.regexp "\\([^/:]*\\)\\(:\\([0-9]+\\)\\)?";
    fields = [ 1, None; 3, Some default_port; ] }

let parse_host host =
  match regexp_match host_regexp host with
  | Some (host :: port :: _) -> host, int_of_string port
  | _ -> error host "Ill formed host"

let url_regexp =
  { regexp = Str.regexp "http://\\([^/:]*\\(:[0-9]+\\)?\\)\\(/.*\\)";
    fields = [ 1, None; 3, None ] }

let parse_url url =
  match regexp_match url_regexp url with
  | Some (host :: path :: _) -> parse_host host, path
  | _ -> error url "Ill formed url"


let get_regexp =
  { regexp = Str.regexp "^[Gg][Ee][Tt][ \t]+\\(.*[^ \t]\\)[ \t]*HTTP/1.[01]\r";
    fields = [ 1, None ] }

let parse_get line =
  match regexp_match get_regexp line with
  | Some (url :: _) -> url
  | _ -> error line "Ill formed request"


let response_regexp =
  { regexp = Str.regexp "^HTTP/1.[01][ \t]+\\([0-9]+\\)[ \t]+\\(.*[^\r]\\)\r";
    fields = [ 1, None; 2, None; ] }

let parse_response line =
  match regexp_match response_regexp line with
  | Some (code :: msg :: _) -> (int_of_string code, msg)
  | _ -> error line "Ill formed response"

let header_regexp =
  { regexp = Str.regexp "\\(.*[^:]\\):\\(.*\\)\r";
    fields = [ 1, None; 2, None ] }

let parse_header line =
  match regexp_match header_regexp line with
  | Some (field :: value :: _) -> field, value
  | _ -> error line "Ill formed header"
