(* D'après le code de Xavier Leroy et Didier Remy : 
     http://gallium.inria.fr/~remy/poly/system/camlunix
*)

open Sys
open Unix 

let () = ignore (signal sigpipe Signal_ignore)

(* ************************** *)
(* Questions préliminaires    *)
(* ************************** *)

(* Question 1 *)

let make_addr host_name port =
  let host_addr =
    try (gethostbyname host_name).h_addr_list.(0)
    with Not_found -> 
      prerr_endline (host_name ^ ": Host not found");
      exit 2
  in
  ADDR_INET (host_addr, port)


(* Question 2 *)

let run_client f addr =
  let sock = socket PF_INET SOCK_STREAM 0 in
  connect sock addr;
  f sock


(* Question 3 *)

let rec remove_headers in_ch =
  match input_line in_ch with
  | "\r" -> ()
  | line -> remove_headers in_ch 

let retransmit in_ch out_ch =
  let buffer_size = 4096 in
  let buffer = String.create buffer_size in
  let rec copy() =
    match input in_ch buffer 0 buffer_size with
    | 0 -> ()
    | n -> 
	ignore (output out_ch buffer 0 n); 
	copy () 
  in
  copy ()


let web_client url =
  let (host_name, port), path = Misc.handle_error Misc.parse_url url in
  let addr = make_addr host_name port in
  run_client 
    (fun fd -> 
      let in_ch = in_channel_of_descr fd in
      let out_ch = out_channel_of_descr fd in
      output_string out_ch ("GET "^path^" HTTP/1.1\r\n");
      output_string out_ch ("Host: "^host_name^"\r\n");
      output_string out_ch "User-Agent: mon programme\r\n";
      output_string out_ch "Connection: close\r\n";
      output_string out_ch "\r\n";
      flush out_ch;
      let response = input_line in_ch in
      begin match Misc.parse_response response with
      | 200, _ ->
          remove_headers in_ch;
          retransmit in_ch Pervasives.stdout
      | err, msg -> failwith (Printf.sprintf "Error %i: %s" err msg)
      end;
      close fd)
    addr

let question3 () = 
  if Array.length Sys.argv < 2 then begin
    prerr_endline ("Usage: "^Sys.argv.(0)^" url");
    exit 1;
  end;
  web_client Sys.argv.(1)

(* let () = *)
(*   handle_unix_error question3 (); *)
(*   exit 0 *)



(* Question 4 *)

let rec restart_on_EINTR f x =
  try f x 
  with Unix_error (EINTR, _, _) -> restart_on_EINTR f x

let run_server service addr =
  let sock = socket PF_INET SOCK_STREAM 0 in
  setsockopt sock SO_REUSEADDR true;
  begin try
    bind sock addr;
    listen sock 10
  with e -> close sock; raise e
  end;
  while true do
    let client_fd, client_addr = restart_on_EINTR accept sock in
    service client_fd client_addr
  done


(* Question 5 *)

let cat_service fd _ =
  Misc.retransmit fd stdout

(* Question 6 *)

let question6 () =
  if Array.length Sys.argv < 2 then begin
    prerr_endline ("Usage: "^Sys.argv.(0)^" port");
    exit 1;
  end;
  let port = int_of_string Sys.argv.(1) in
  let addr = ADDR_INET (inet_addr_any, port) in
  run_server cat_service addr

(* let () = handle_unix_error question6 () *)




(* ************************** *)
(* Proxy transparent          *)
(* ************************** *)

(* Question 7 *)

let proxy_service_v0 fd _ =
  let in_ch = in_channel_of_descr fd in
  let out_ch = out_channel_of_descr fd in
  remove_headers in_ch;
  output_string out_ch "HTTP/1.1 403 Forbidden\r\n";
  output_string out_ch "Connection: close\r\n";
  output_string out_ch "Content-type: text/html\r\n";
  output_string out_ch "\r\n";
  output_string out_ch "<html><head><title>Erreur !</title></head>\r\n";
  output_string out_ch "<body>Je ne veux pas aller chercher cette page.</body></html>\r\n";
  flush out_ch;
  close fd

let question7 () =
  if Array.length Sys.argv < 2 then begin
    prerr_endline ("Usage: "^Sys.argv.(0)^" port");
    exit 1;
  end;
  let port = int_of_string Sys.argv.(1) in
  let addr = ADDR_INET (inet_addr_any, port) in
  run_server proxy_service_v0 addr

(* let () = handle_unix_error question7 () *)
 
(* Question 8 *)
let rec retransmit_headers client server =
  match input_line client with
  | "\r" -> 
      output_string server "\r\n";
      flush server
  | line -> 
      let field, _ = Misc.parse_header line in
      begin match String.lowercase field with
      | "proxy-connection" | "connection" -> 
	  print_endline ("<<< "^ line )
      | _ -> 
	  print_endline (">>> "^ line );
	  output_string server (line^"\n")
      end;
      retransmit_headers client server
  

let error_502 fd msg =
  let out_ch = out_channel_of_descr fd in
  output_string out_ch "HTTP/1.1 502 Bad Gateway \r\n";
  output_string out_ch "Connection: close\r\n";
  output_string out_ch "Content-type: text/html\r\n";
  output_string out_ch "\r\n";
  output_string out_ch "<html><head><title>Erreur !</title></head>\r\n";
  output_string out_ch ("<body>"^msg^"</body></html>\r\n");
  flush out_ch

let proxy_service fd_client _ =
  try 
    let client = in_channel_of_descr fd_client in
    let url = Misc.parse_get (input_line client) in
    let (host_name, port), path = Misc.parse_url url in
    let addr = make_addr host_name port in
    run_client 
      (fun fd_server ->
	let server = out_channel_of_descr fd_server in
	output_string server ("GET "^path^" HTTP/1.1\r\n");
	output_string server "Connection: close\r\n";
	retransmit_headers client server;
	Misc.retransmit fd_server fd_client;
	close fd_server;
	close fd_client)
      addr
  with
  | Misc.Error msg -> 
      prerr_endline msg; 
      error_502 fd_client msg;
      close fd_client
  | Unix_error (err, "connect", _)  ->
      let msg = error_message err in
      prerr_endline msg; 
      error_502 fd_client msg;
      close fd_client
  | _ -> ()

let question8 () =
  if Array.length Sys.argv < 2 then begin
    prerr_endline ("Usage: "^Sys.argv.(0)^" port");
    exit 1;
  end;
  let port = int_of_string Sys.argv.(1) in
  let addr = ADDR_INET (inet_addr_any, port) in
  run_server proxy_service addr

let () = handle_unix_error question8 ()
